#!/usr/bin/env python

"""
Simplify AST structures for easier type propagation and analysis. The code in
this module processes AST trees originating from the compiler module and
produces a result tree consisting of instruction-oriented program nodes.

Copyright (C) 2006 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

--------

To use this module, the easiest approach is to use the simplify function:

simplify(filename)

The more complicated approach involves first instantiating a Simplifier object:

simplifier = Simplifier()

Then, applying the simplifier to an AST tree:

module = compiler.parseFile(filename)
simplifier.process(module)
"""

from simplified import *
import compiler.ast

class Simplifier(Visitor):

    """
    A simplifying visitor for AST nodes.

    Covered: And, AssAttr, AssList, AssName, AssTuple, Assign, AugAssign, Break,
             CallFunc, Class, Compare, Const, Continue, Dict, Discard, For,
             From, Function, Getattr, Global, If, Import, Invert, Keyword,
             Lambda, List, Module, Name, Not, Or, Pass, Raise, Return, Slice,
             Stmt, Subscript, TryExcept, TryFinally, Tuple, While, UnaryAdd,
             UnarySub.

    Missing: Add, Assert, Backquote, Bitand, Bitor, Bitxor, Decorators, Div,
             Ellipsis, Exec, FloorDiv, LeftShift, ListComp, ListCompFor,
             ListCompIf, Mod, Mul, Power, Print, Printnl, RightShift, Sliceobj,
             Sub, Yield.
    """

    def __init__(self, builtins=0):

        """
        Initialise the simplifier with the optional 'builtins' parameter
        indicating whether the module contains the built-in classes and
        functions.
        """

        Visitor.__init__(self)
        self.subprograms = []           # Subprograms outside the tree.
        self.structures = []            # Structures/classes.
        self.constants = {}             # Constants.
        self.current_subprograms = []   # Current subprograms being processed.
        self.builtins = builtins        # Whether the builtins are being processed.

        # For compiler package mechanisms.

        self.visitor = self

    def process(self, node):
        result = self.dispatch(node)
        result.simplifier = self
        return result

    def dispatch_or_none(self, node, *args):
        if node is not None:
            return self.dispatch(node, *args)
        else:
            return LoadName(name="None")

    # Placeholder or deletion transformations.

    def visitStmt(self, stmt):
        return self.dispatches(stmt.nodes)

    def visitPass(self, pass_):
        return Pass(pass_)

    def visitDiscard(self, discard):
        return self.dispatch(discard.expr)

    # Relatively trivial transformations.

    def visitModule(self, module):

        """
        Process the given 'module', producing a Module object which contains the
        resulting program nodes.
        """

        result = Module(module)
        module_code = self.dispatch(module.node)

        # NOTE: Constant initialisation necessary for annotation but perhaps
        # NOTE: redundant in the program.

        init_code = []
        for value, constant in self.constants.items():
            init_code.append(
                StoreAttr(module,
                    lvalue=LoadRef(module, ref=constant),
                    name="__class__",
                    expr=LoadName(module, name=constant.typename)
                    )
                )

        # NOTE: Hack to ensure correct initialisation of constants.

        if self.builtins:
            result.code = module_code + init_code
        else:
            result.code = init_code + module_code 
        return result

    def visitGetattr(self, getattr):
        result = LoadAttr(getattr,
            name=getattr.attrname,
            expr=self.dispatch(getattr.expr)
            )
        return result

    def visitKeyword(self, keyword):
        result = Keyword(keyword,
            name=keyword.name,
            expr=self.dispatch(keyword.expr)
            )
        return result

    def visitGlobal(self, global_):
        result = Global(global_,
            names=global_.names
            )
        return result

    def visitImport(self, import_):
        result = Assign(import_)
        code = []
        for path, alias in import_.names:
            importer = Import(import_, name=path)
            top = alias or path.split(".")[0]
            code.append(StoreName(import_, expr=importer, name=top))
        result.code = code
        return result

    def visitFrom(self, from_):
        result = Assign(from_)
        code = []
        code.append(StoreTemp(from_, expr=Import(from_, name=from_.modname)))
        for name, alias in from_.names:
            code.append(
                StoreName(from_,
                    expr=LoadAttr(from_,
                        expr=LoadTemp(from_),
                        name=name),
                    name=(alias or name)
                    )
                )
        result.code = code
        return result

    def visitName(self, name):
        result = LoadName(name, name=name.name)
        return result

    def visitConst(self, const):
        if not self.constants.has_key(const.value):
            self.constants[const.value] = Constant(const, name=repr(const.value), value=const.value)
        result = LoadRef(const, ref=self.constants[const.value])
        return result

    def visitReturn(self, return_):
        result = Return(return_,
            expr=self.dispatch(return_.value)
            )
        return result

    def visitBreak(self, break_):
        result = Return(break_)
        return result

    def visitContinue(self, continue_):
        result = InvokeBlock(continue_,
            expr=LoadRef(continue_, ref=self.current_subprograms[-1])
            )
        return result

    def visitRaise(self, raise_):
        result = Raise(raise_, expr=self.dispatch(raise_.expr1), traceback=None)
        if raise_.expr2 is not None:
            result.args = [self.dispatch(raise_.expr2)]
        if raise_.expr3 is not None:
            result.traceback = self.dispatch(raise_.expr3)
        return result

    def _visitBuiltin(self, builtin, name):
        result = Invoke(builtin, expr=LoadName(builtin, name=name), args=self.dispatches(builtin.nodes), star=None, dstar=None)
        return result

    def visitTuple(self, tuple):
        return self._visitBuiltin(tuple, "tuple")

    def visitList(self, list):
        return self._visitBuiltin(list, "list")

    def visitDict(self, dict):
        result = Invoke(dict, expr=LoadName(dict, name="dict"), star=None, dstar=None)
        args = []
        for key, value in dict.items:
            tuple = Invoke(dict, expr=LoadName(dict, name="tuple"), star=None, dstar=None)
            tuple.args = [self.dispatch(key), self.dispatch(value)]
            args.append(tuple)
        result.args = args
        return result

    # Logical and comparison operations plus chained statements.

    def visitIf(self, if_):

        """
        Make conditionals for each test from an 'if_' AST node, adding the body
        and putting each subsequent test as part of the conditional's else
        section.

        Convert...

        If (test/body)
           (test/body)
           ...
           (else/body)

        ...to:

        Conditional (test) -> (body)
                    (else) -> Conditional (test) -> (body)
                                          (else) -> ...
        """


        results = nodes = []

        # Produce something like...
        # expr.__true__() ? body

        for compare, stmt in if_.tests:
            test = Conditional(if_,
                test=Invoke(if_,
                    expr=LoadAttr(if_,
                        expr=self.dispatch(compare),
                        name="__true__"
                        ),
                    args=[],
                    star=None,
                    dstar=None)
                )
            test.body = self.dispatch(stmt)
            nodes.append(test)
            nodes = test.else_ = []

        # Add the compound statement from any else clause to the end.

        if if_.else_ is not None:
            nodes += self.dispatch(if_.else_)

        result = results[0]
        return result

    def visitTryExcept(self, tryexcept):

        """
        Make conditionals for each handler associated with a 'tryexcept' node.

        Convert...

        TryExcept (body)
                  (else)
                  (spec/assign/stmt)
                  ...

        ...to:

        Try (body)
            (else)
            (handler) -> Conditional (test) -> (stmt)
                                     (body) -> ...
                                     (else) -> Conditional (test) -> (stmt)
                                                           (body) -> ...
                                                           (else) -> ...
        """
                  
        result = Try(tryexcept, body=[], else_=[], finally_=[])

        if tryexcept.body is not None:
            result.body = self.dispatch(tryexcept.body)
        if tryexcept.else_ is not None:
            result.else_ = self.dispatch(tryexcept.else_)

        results = nodes = []
        for spec, assign, stmt in tryexcept.handlers:

            # If no specification exists, produce an unconditional block.

            if spec is None:
                nodes += self.dispatch(stmt)

            # Produce something like...
            # isinstance(<exc>, <spec>)

            else:
                new_spec = self.dispatch(spec)
                test = Conditional(tryexcept,
                    test=Invoke(tryexcept,
                        expr=LoadName(tryexcept, name="isinstance"),
                        args=[LoadExc(tryexcept), new_spec],
                        star=None,
                        dstar=None)
                    )
                test.body = []

                if assign is not None:
                    test.body.append(
                        Assign(tryexcept,
                            code=[StoreTemp(expr=LoadExc(tryexcept)), self.dispatch(assign), ReleaseTemp(tryexcept)])
                        )

                # Always return from conditional sections.

                test.body += self.dispatch(stmt) + [Return(tryexcept)]
                nodes.append(test)
                nodes = test.else_ = []

        # Add a raise operation to deal with unhandled exceptions.

        nodes.append(
            Raise(tryexcept,
                expr=LoadExc(tryexcept))
            )

        result.handler = results
        return result

    comparison_methods = {
        "==" : "__eq__", "!=" : "__ne__", "<" : "__lt__", "<=" : "__le__",
        ">=" : "__ge__", ">" : "__gt__", "is" : None, "is not" : None
        }

    def visitCompare(self, compare):

        """
        Make a subprogram for the 'compare' node and record its contents inside
        the subprogram. Convert...

        Compare (expr)
                (name/node)
                ...

        ...to:

        Subprogram -> Conditional (test) -> (body)
                                  (else) -> Conditional (test) -> (body)
                                                        (else) -> ...
        """

        subprogram = Subprogram(compare, name=None, acquire_locals=1, returns_value=1, params=[], star=None, dstar=None)
        self.current_subprograms.append(subprogram)

        # In the subprogram, make instructions which invoke a method on the
        # first operand of each operand pair and, if appropriate, return with
        # the value from that method.

        last = compare.ops[-1]
        previous = self.dispatch(compare.expr)
        results = nodes = []

        for op in compare.ops:
            op_name, node = op
            expr = self.dispatch(node)

            # Identify the operation and produce the appropriate method call.

            method_name = self.comparison_methods[op_name]
            if method_name:
                invocation = Invoke(compare, expr=LoadAttr(compare, expr=previous, name=method_name), args=[expr], star=None, dstar=None)
            elif op_name == "is":
                invocation = Invoke(compare, expr=LoadName(compare, name="__is__"), args=[previous, expr], star=None, dstar=None)
            elif op_name == "is not":
                invocation = Not(compare, expr=Invoke(compare, expr=LoadName(compare, name="__is__"), args=[previous, expr], star=None, dstar=None))
            else:
                raise NotImplementedError, op_name
            nodes.append(StoreTemp(compare, expr=invocation))

            # Return from the subprogram where the test is not satisfied.

            if op is not last:
                test = Conditional(compare, test=Not(compare, expr=LoadTemp(compare)), body=[Return(compare, expr=LoadTemp(compare))])
                nodes.append(test)

                # Put subsequent operations in the else section of this conditional.

                nodes = test.else_ = [ReleaseTemp(compare)]

            # For the last operation, return the result.

            else:
                nodes.append(Return(compare, expr=LoadTemp(compare)))

            previous = expr

        # Finish the subprogram definition.

        subprogram.code = results

        self.current_subprograms.pop()
        self.subprograms.append(subprogram)

        # Make an invocation of the subprogram.

        result = InvokeBlock(compare, produces_result=1)
        result.expr = LoadRef(compare, ref=subprogram)
        return result

    def visitAnd(self, and_):

        """
        Make a subprogram for the 'and_' node and record its contents inside the
        subprogram. Convert...

        And (test)
            (test)
            ...

        ...to:

        Subprogram -> Conditional (test) -> Return ...
                                  (else) -> Conditional (test) -> Return ...
                                                        (else) -> ...
        """

        subprogram = Subprogram(and_, name=None, acquire_locals=1, returns_value=1, params=[], star=None, dstar=None)
        self.current_subprograms.append(subprogram)

        # In the subprogram, make instructions which store each operand, test
        # for each operand's truth status, and if appropriate return from the
        # subprogram with the value of the operand.

        last = and_.nodes[-1]
        results = nodes = []

        for node in and_.nodes:
            expr = self.dispatch(node)

            # Return from the subprogram where the test is not satisfied.

            if node is not last:
                nodes.append(StoreTemp(and_, expr=expr))
                invocation = Invoke(and_, expr=LoadAttr(and_, expr=LoadTemp(and_), name="__true__"), args=[], star=None, dstar=None)
                test = Conditional(and_, test=Not(and_, expr=invocation), body=[Return(and_, expr=LoadTemp(and_))])
                nodes.append(test)

                # Put subsequent operations in the else section of this conditional.

                nodes = test.else_ = [ReleaseTemp(and_)]

            # For the last operation, return the result.

            else:
                nodes.append(Return(and_, expr=expr))

        # Finish the subprogram definition.

        subprogram.code = results

        self.current_subprograms.pop()
        self.subprograms.append(subprogram)

        # Make an invocation of the subprogram.

        result = InvokeBlock(and_, produces_result=1)
        result.expr = LoadRef(and_, ref=subprogram)
        return result

    def visitOr(self, or_):

        """
        Make a subprogram for the 'or_' node and record its contents inside the
        subprogram. Convert...

        Or (test)
           (test)
           ...

        ...to:

        Subprogram -> Conditional (test) -> Return ...
                                  (else) -> Conditional (test) -> Return ...
                                                        (else) -> ...
        """

        subprogram = Subprogram(or_, name=None, acquire_locals=1, returns_value=1, params=[], star=None, dstar=None)
        self.current_subprograms.append(subprogram)

        # In the subprogram, make instructions which store each operand, test
        # for each operand's truth status, and if appropriate return from the
        # subprogram with the value of the operand.

        last = or_.nodes[-1]
        results = nodes = []

        for node in or_.nodes:
            expr = self.dispatch(node)

            # Return from the subprogram where the test is satisfied.

            if node is not last:
                nodes.append(StoreTemp(or_, expr=expr))
                invocation = Invoke(or_, expr=LoadAttr(or_, expr=LoadTemp(or_), name="__true__"), args=[], star=None, dstar=None)
                test = Conditional(or_, test=invocation, body=[Return(or_, expr=LoadTemp(or_))])
                nodes.append(test)

                # Put subsequent operations in the else section of this conditional.

                nodes = test.else_ = [ReleaseTemp(or_)]

            # For the last operation, return the result.

            else:
                nodes.append(
                    Return(or_, expr=expr)
                    )

        # Finish the subprogram definition.

        subprogram.code = results

        self.current_subprograms.pop()
        self.subprograms.append(subprogram)

        # Make an invocation of the subprogram.

        result = InvokeBlock(or_, produces_result=1)
        result.expr = LoadRef(or_, ref=subprogram)
        return result

    def visitNot(self, not_):
        result = Not(not_, expr=Invoke(not_, expr=LoadAttr(not_, expr=self.dispatch(not_.expr), name="__true__"), args=[], star=None, dstar=None))
        return result

    # Operators.

    def visitUnaryAdd(self, unaryadd):
        return Invoke(unaryadd, expr=LoadAttr(unaryadd, expr=self.dispatch(unaryadd.expr), name="__pos__"), args=[], star=None, dstar=None)

    def visitUnarySub(self, unarysub):
        return Invoke(unarysub, expr=LoadAttr(unarysub, expr=self.dispatch(unarysub.expr), name="__neg__"), args=[], star=None, dstar=None)

    def visitInvert(self, invert):
        return Invoke(invert, expr=LoadAttr(invert, expr=self.dispatch(invert.expr), name="__invert__"), args=[], star=None, dstar=None)

    def visitAdd(self, add):

        """
        Emulate the current mechanisms by producing nodes as follows:

        Try (body) -> x.__add__(y)
            (else)
            (handler) -> Conditional (test) -> isinstance(exc, TypeError)
                                     (body) -> y.__radd__(x)
                                     (else)
        """

        result = Try(add,
            body=[
                Invoke(add,
                    expr=LoadAttr(add, expr=self.dispatch(add.left), name="__add__"),
                    args=[self.dispatch(add.right)],
                    star=None,
                    dstar=None)
                ],
            else_=[],
            finally_=[],
            handler=[
                Conditional(add,
                    test=Invoke(add,
                        expr=LoadName(add, name="isinstance"),
                        args=[LoadExc(add), LoadName(add, name="TypeError")],
                        star=None,
                        dstar=None),
                    body=[
                        Invoke(add,
                            expr=LoadAttr(add, expr=self.dispatch(add.right), name="__radd__"),
                            args=[self.dispatch(add.left)],
                            star=None,
                            dstar=None)
                        ],
                    else_=[]
                    )
                ]
            )

        return result

    # Assignments.

    augassign_methods = {
        "+=" : "__iadd__", "-=" : "__isub__", "*=" : "__imul__", "/=" : "__idiv__",
        "%=" : "__imod__", "**=" : "__ipow__", "<<=" : "__ilshift__", ">>=" : "__irshift__",
        "&=" : "__iand__", "^=" : "__ixor__", "|=" : "__ior__"
        }

    def visitAugAssign(self, augassign):
        result = Assign(augassign)
        expr = self.dispatch(augassign.expr)

        # Simple augmented assignment: name += expr

        if isinstance(augassign.node, compiler.ast.Name):
            result.code = [
                StoreTemp(augassign,
                    expr=Invoke(augassign,
                        args=[expr],
                        star=None,
                        dstar=None,
                        expr=LoadAttr(augassign,
                            expr=self.dispatch(augassign.node),
                            name=self.augassign_methods[augassign.op]
                            )
                        )
                    ),
                StoreName(augassign,
                    expr=LoadTemp(augassign),
                    name=augassign.node.name),
                ReleaseTemp(augassign)
                ]

        # Complicated augmented assignment: lvalue.attr += expr

        elif isinstance(augassign.node, compiler.ast.Getattr):

            # <lvalue> -> setattr(<lvalue>, getattr(<lvalue>, "attr").__xxx__(expr))

            result.code = [
                StoreTemp(augassign,
                    index="expr",
                    expr=self.dispatch(augassign.node.expr)
                    ),
                StoreTemp(augassign,
                    expr=Invoke(augassign,
                        args=[expr], star=None, dstar=None,
                        expr=LoadAttr(augassign,
                            expr=LoadAttr(augassign,
                                expr=LoadTemp(augassign, index="expr"),
                                name=augassign.node.attrname
                                ),
                            name=self.augassign_methods[augassign.op]
                            )
                        )
                    ),
                StoreAttr(augassign,
                    expr=LoadTemp(augassign),
                    lvalue=LoadTemp(augassign, index="expr"),
                    name=augassign.node.attrname
                    ),
                ReleaseTemp(augassign, index="expr"),
                ReleaseTemp(augassign)
                ]

        # Complicated augassign using slices: lvalue[lower:upper] += expr

        elif isinstance(augassign.node, compiler.ast.Slice):

            # <lvalue>, <lower>, <upper> -> <lvalue>.__setslice__(<lower>, <upper>, <lvalue>.__getslice__(<lower>, <upper>).__xxx__(expr))

            result.code = [
                StoreTemp(augassign,
                    index="expr",
                    expr=self.dispatch(augassign.node.expr)
                    ),
                StoreTemp(augassign,
                    index="lower",
                    expr=self.dispatch_or_none(augassign.node.lower)
                    ),
                StoreTemp(augassign,
                    index="upper",
                    expr=self.dispatch_or_none(augassign.node.upper)
                    ),
                StoreTemp(augassign,
                    expr=Invoke(augassign,
                        args=[expr], star=None, dstar=None,
                        expr=LoadAttr(augassign,
                            expr=self._visitSlice(
                                augassign.node,
                                LoadTemp(augassign, index="expr"),
                                LoadTemp(augassign, index="lower"),
                                LoadTemp(augassign, index="upper"),
                                "OP_APPLY"),
                            name=self.augassign_methods[augassign.op]
                            )
                        )
                    ),
                self._visitSlice(
                    augassign.node,
                    LoadTemp(augassign, index="expr"),
                    LoadTemp(augassign, index="lower"),
                    LoadTemp(augassign, index="upper"),
                    "OP_ASSIGN",
                    LoadTemp(augassign)
                    ),
                ReleaseTemp(augassign, index="expr"),
                ReleaseTemp(augassign, index="lower"),
                ReleaseTemp(augassign, index="upper"),
                ReleaseTemp(augassign)
                ]

        # Complicated augassign using subscripts: lvalue[subs] += expr

        elif isinstance(augassign.node, compiler.ast.Subscript):

            # <lvalue>, <subs> -> <lvalue>.__setitem__(<subs>, <lvalue>.__getitem__(<subs>).__xxx__(expr))

            result.code = [
                StoreTemp(augassign, index="expr", expr=self.dispatch(augassign.node.expr)),
                StoreTemp(augassign, index="subs", expr=self._visitSubscriptSubs(augassign.node.subs)),
                StoreTemp(augassign, 
                    expr=Invoke(augassign, 
                        args=[expr], star=None, dstar=None,
                        expr=LoadAttr(augassign, 
                            expr=self._visitSubscript(
                                augassign.node,
                                LoadTemp(augassign, index="expr"),
                                LoadTemp(augassign, index="subs"),
                                "OP_APPLY"
                                ),
                            name=self.augassign_methods[augassign.op]
                            )
                        )
                    ),
                self._visitSubscript(
                    augassign.node,
                    LoadTemp(augassign, index="expr"),
                    LoadTemp(augassign, index="subs"),
                    "OP_ASSIGN",
                    LoadTemp(augassign)
                    ),
                ReleaseTemp(augassign, index="expr"),
                ReleaseTemp(augassign, index="subs"),
                ReleaseTemp(augassign)
                ]

        else:
            raise NotImplementedError, augassign.node.__class__

        return result

    def visitAssign(self, assign):
        result = Assign(assign)
        store = StoreTemp(assign, expr=self.dispatch(assign.expr))
        release = ReleaseTemp(assign)
        result.code = [store] + self.dispatches(assign.nodes, 0) + [release]
        return result

    def visitAssList(self, asslist, in_sequence=0):
        if not in_sequence:
            expr = LoadTemp(asslist)
        else:
            expr = Invoke(asslist, expr=LoadAttr(asslist, expr=LoadTemp(asslist), name="next"), star=None, dstar=None, args=[])
        result = Assign(asslist)
        store = StoreTemp(asslist, expr=Invoke(asslist, expr=LoadAttr(asslist, name="__iter__", expr=expr), star=None, dstar=None, args=[]))
        release = ReleaseTemp(asslist)
        result.code = [store] + self.dispatches(asslist.nodes, 1) + [release]
        return result

    visitAssTuple = visitAssList

    def _visitAssNameOrAttr(self, node, in_sequence):
        if not in_sequence:
            return LoadTemp(node)
        else:
            return Invoke(node, expr=LoadAttr(node, expr=LoadTemp(node), name="next"), star=None, dstar=None, args=[])

    def visitAssName(self, assname, in_sequence=0):
        expr = self._visitAssNameOrAttr(assname, in_sequence)
        result = StoreName(assname, name=assname.name, expr=expr)
        return result

    def visitAssAttr(self, assattr, in_sequence=0):
        expr = self._visitAssNameOrAttr(assattr, in_sequence)
        lvalue = self.dispatch(assattr.expr)
        result = StoreAttr(assattr, name=assattr.attrname, lvalue=lvalue, expr=expr)
        return result

    def _visitSlice(self, slice, expr, lower, upper, flags, value=None):
        if flags == "OP_ASSIGN":
            args = [value]
            result = Invoke(slice, expr=LoadAttr(slice, expr=expr, name="__setslice__"), star=None, dstar=None, args=[])
        elif flags == "OP_APPLY":
            args = []
            result = Invoke(slice, expr=LoadAttr(slice, expr=expr, name="__getslice__"), star=None, dstar=None, args=[])
        elif flags == "OP_DELETE":
            args = []
            result = Invoke(slice, expr=LoadAttr(slice, expr=expr, name="__delslice__"), star=None, dstar=None, args=[])
        else:
            raise NotImplementedError, flags

        # Add the dimensions.

        args.insert(0, lower)
        args.insert(1, upper)

        result.original = slice
        result.args = args
        return result

    def visitSlice(self, slice, in_sequence=0):
        return self._visitSlice(slice, self.dispatch(slice.expr), self.dispatch_or_none(slice.lower),
            self.dispatch_or_none(slice.upper), slice.flags, self._visitAssNameOrAttr(slice, in_sequence))

    def _visitSubscript(self, subscript, expr, subs, flags, value=None):
        if flags == "OP_ASSIGN":
            args = [value]
            result = Invoke(subscript, expr=LoadAttr(subscript, expr=expr, name="__setitem__"), star=None, dstar=None, args=[])
        elif flags == "OP_APPLY":
            args = []
            result = Invoke(subscript, expr=LoadAttr(subscript, expr=expr, name="__getitem__"), star=None, dstar=None, args=[])
        elif flags == "OP_DELETE":
            args = []
            result = Invoke(subscript, expr=LoadAttr(subscript, expr=expr, name="__delitem__"), star=None, dstar=None, args=[])
        else:
            raise NotImplementedError, flags

        # Add the dimensions.

        args.insert(0, subs)

        result.original = subscript
        result.args = args
        return result

    def _visitSubscriptSubs(self, subs):
        if len(subs) == 1:
            return self.dispatch(subs[0])
        else:
            return Invoke(subs, expr=LoadName(subs, name="tuple"), args=self.dispatches(subs), star=None, dstar=None)

    def visitSubscript(self, subscript, in_sequence=0):
        return self._visitSubscript(
            subscript, self.dispatch(subscript.expr), self._visitSubscriptSubs(subscript.subs), subscript.flags,
            self._visitAssNameOrAttr(subscript, in_sequence)
            )

    # Invocation and subprogram transformations.

    def visitClass(self, class_):
        structure = Class(class_, name=class_.name, bases=self.dispatches(class_.bases))
        self.structures.append(structure)

        # Make a subprogram which initialises the class structure.

        subprogram = Subprogram(class_, name=None, structure=structure, params=[], star=None, dstar=None)
        self.current_subprograms.append(subprogram)

        # The class is initialised using the code found inside.

        subprogram.code = self.dispatch(class_.code) + [Return(class_)]

        self.current_subprograms.pop()
        self.subprograms.append(subprogram)

        # Make a definition of the class associating it with a name.

        result = Assign(class_,
            code=[
                StoreName(class_,
                    name=class_.name,
                    expr=LoadRef(class_, ref=structure)
                    ),
                Invoke(class_,
                    args=[], star=None, dstar=None,
                    expr=LoadRef(class_, ref=subprogram)
                    )
                ]
            )
        return result

    def _visitFunction(self, function, subprogram):

        """
        A common function generator which transforms the given 'function' node
        and initialises the given 'subprogram' appropriately.
        """

        # Discover star and dstar parameters.

        if function.flags & 4 != 0:
            has_star = 1
        else:
            has_star = 0
        if function.flags & 8 != 0:
            has_dstar = 1
        else:
            has_dstar = 0

        # Discover the number of defaults and positional parameters.

        ndefaults = len(function.defaults)
        npositional = len(function.argnames) - has_star - has_dstar

        # Produce star and dstar parameters with appropriate defaults.

        if has_star:
            star = (function.argnames[npositional], Invoke(function, expr=LoadName(function, name="list"), args=[], star=None, dstar=None))
        else:
            star = None
        if has_dstar:
            dstar = (function.argnames[npositional + has_star], Invoke(function, expr=LoadName(function, name="dict"), args=[], star=None, dstar=None))
        else:
            dstar = None

        params = []
        for i in range(0, npositional - ndefaults):
            params.append((function.argnames[i], None))

        # NOTE: Fix/process defaults.

        for i in range(0, ndefaults):
            default = function.defaults[i]
            if default is not None:
                params.append((function.argnames[npositional - ndefaults + i], self.dispatch(default)))
            else:
                params.append((function.argnames[npositional - ndefaults + i], default))

        subprogram.params = params
        subprogram.star = star
        subprogram.dstar = dstar
        self.subprograms.append(subprogram)

    def visitFunction(self, function):

        # Make a subprogram for the function and record it outside the main
        # tree.

        subprogram = Subprogram(function, name=function.name, acquire_locals=0, returns_value=1, star=None, dstar=None)
        self.current_subprograms.append(subprogram)
        subprogram.code = self.dispatch(function.code) + [Return(function)]
        self.current_subprograms.pop()
        self._visitFunction(function, subprogram)

        # Make a definition of the function associating it with a name.

        result = StoreName(function, name=function.name, expr=LoadRef(function, ref=subprogram))
        return result

    def visitLambda(self, lambda_):

        # Make a subprogram for the function and record it outside the main
        # tree.

        subprogram = Subprogram(lambda_, name=None, acquire_locals=0, returns_value=1, star=None, dstar=None)
        self.current_subprograms.append(subprogram)
        subprogram.code = [Return(lambda_, expr=self.dispatch(lambda_.code))]
        self.current_subprograms.pop()
        self._visitFunction(lambda_, subprogram)

        # Get the subprogram reference to the lambda.

        return LoadRef(lambda_, ref=subprogram)

    def visitCallFunc(self, callfunc):
        result = InvokeFunction(callfunc, star=None, dstar=None)
        result.args = self.dispatches(callfunc.args)
        if callfunc.star_args is not None:
            result.star = self.dispatch(callfunc.star_args)
        if callfunc.dstar_args is not None:
            result.dstar = self.dispatch(callfunc.dstar_args)
        result.expr = self.dispatch(callfunc.node)
        return result

    def visitWhile(self, while_):

        """
        Make a subprogram for the 'while' node and record its contents inside the
        subprogram. Convert...

        While (test) -> (body)
              (else)

        ...to:

        Subprogram -> Conditional (test) -> (body) -> Invoke subprogram
                                  (else) -> Conditional (test) -> Return ...
                                                        (else) -> ...
        """

        subprogram = Subprogram(while_, name=None, acquire_locals=1, returns_value=0, params=[], star=None, dstar=None)
        self.current_subprograms.append(subprogram)

        # Include a conditional statement in the subprogram.

        test = Conditional(while_, else_=[])
        test.test = Invoke(while_, expr=LoadAttr(while_, expr=self.dispatch(while_.test), name="__true__"), args=[], star=None, dstar=None)

        # Inside the conditional, add a recursive invocation to the subprogram
        # if the test condition was satisfied.

        continuation = InvokeBlock(while_)
        continuation.expr = LoadRef(while_, ref=subprogram)

        # Return within the main section of the loop.

        test.body = self.dispatch(while_.body) + [continuation, Return(while_)]

        # Provide the else section, if present, along with an explicit return.

        if while_.else_ is not None:
            test.else_ = self.dispatch(while_.else_) + [Return(while_)]

        # Finish the subprogram definition.

        subprogram.code = [test]

        self.current_subprograms.pop()
        self.subprograms.append(subprogram)

        # Make an invocation of the subprogram.

        result = InvokeBlock(while_)
        result.expr = LoadRef(while_, ref=subprogram)
        return result

    def visitFor(self, for_):

        """
        Make a subprogram for the 'for_' node and record its contents inside the
        subprogram. Convert...

        For (assign)
            (body)
            (else)

        ...to:

        Assign (assign #1)
        Invoke -> Subprogram -> Try (body) -> (assign #2)
                                              (body)
                                              Invoke subprogram
                                    (handler) -> ...
                                    (else) -> ...
        """

        subprogram = Subprogram(for_, name=None, acquire_locals=1, returns_value=0, params=[], star=None, dstar=None)
        self.current_subprograms.append(subprogram)

        # Always return from conditional sections/subprograms.

        if for_.else_ is not None:
            else_stmt = self.dispatch(for_.else_) + [Return(for_)]
        else:
            else_stmt = [Return(for_)]

        # Wrap the assignment in a try...except statement.

        try_except = Try(for_, body=[], else_=[], finally_=[])
        test = Conditional(for_, 
            test=Invoke(for_, expr=LoadName(for_, name="isinstance"), args=[LoadExc(for_), LoadName(for_, name="StopIteration")], star=None, dstar=None),
            body=else_stmt,
            else_=[Raise(for_, expr=LoadExc(for_))])
        try_except.handler = [test]

        assign = Assign(for_,
            code=[
                StoreTemp(for_, expr=Invoke(for_, expr=LoadAttr(for_, expr=LoadTemp(for_), name="next"), args=[], star=None, dstar=None)),
                self.dispatch(for_.assign),
                ReleaseTemp(for_)
                ])

        # Inside the conditional, add a recursive invocation to the subprogram
        # if the test condition was satisfied.

        continuation = InvokeBlock(for_)
        continuation.expr = LoadRef(for_, ref=subprogram)
        try_except.body = [assign] + self.dispatch(for_.body) + [continuation]
        subprogram.code = [try_except, Return(for_)]

        # Finish the subprogram definition.

        self.subprograms.append(subprogram)
        self.current_subprograms.pop()

        # Obtain an iterator for the sequence involved.
        # Then, make an invocation of the subprogram.

        result = Assign(for_)
        result.code = [
            StoreTemp(for_, expr=Invoke(for_, expr=LoadAttr(for_, name="__iter__", expr=self.dispatch(for_.list)), args=[], star=None, dstar=None)),
            InvokeBlock(for_, expr=LoadRef(for_, ref=subprogram)),
            ReleaseTemp(for_)
            ]
        return result

    # Exception node transformations.

    def visitTryFinally(self, tryfinally):
        result = Try(tryfinally, body=[], else_=[], finally_=[])
        if tryfinally.body is not None:
            result.body = self.dispatch(tryfinally.body)
        if tryfinally.final is not None:
            result.finally_ = self.dispatch(tryfinally.final)
        return result

# Convenience functions.

def simplify(filename, builtins=0):

    """
    Simplify the module stored in the file with the given 'filename'. If the
    optional 'builtins' parameter is set to a true value (the default being a
    false value), then the module is considered as the builtins module.
    """

    simplifier = Simplifier(builtins)
    module = compiler.parseFile(filename)
    simplified = simplifier.process(module)
    compiler.misc.set_filename(filename, module)
    return simplified

# vim: tabstop=4 expandtab shiftwidth=4
